package estimateur.modele;

import java.util.Collections;
import java.util.Vector;

/**
 * Description d'une source de financement
 * 
 * Étant donnée la nature d'une source de financement, 5 quantités suffisent
 * pour la décrire :
 * - 4 flottants pour traduire l'imprécision avec un nombre flou
 * 	(bornes du support et du noyau)
 * - l'incertitude, ici élément d'une énumération (5 niveaux gérés)
 * 
 * Un financement aura donc toujours une représentation du type "barre en 0" et
 * trapèze.
 */
public class Financement {

	/*
	 * Les flottants qui décrivent la source
	 */
	private Double supportInf, supportSup, noyauInf, noyauSup;
	// l'incertitude de la source
	Incertitude incertitude;
	
	/*
	 * Description textuelle de la source de financement
	 */
	private String desc;

	/**
	 * Construction d'une source de financement
	 * @param supportInf: borne inférieure du support
	 * @param supportSup: borne supérieure du support
	 * @param noyauInf: borne inférieure du noyau
	 * @param noyauSup: borne supérieure du noyau
	 * @param incertitude: niveau d'incertitude de la source de financement
	 * @param description: description textuelle de la source de financement
	 */
	public Financement(Double supportInf, Double supportSup,
			Double noyauInf, Double noyauSup,
			Incertitude incertitude,
			String description)
			throws IntervalleMalDefiniException, NombreFlouMalDefiniException {
		
		/*
		 * On prend garde à ce que la définition soit cohérente
		 */
		if ( supportInf > supportSup) {
			throw new IntervalleMalDefiniException(
			"Ensemble des valeurs possibles mal défini\n" +
			"Veuillez convenablement spécifier les bornes inférieure et supérieure");
		}
		if ( noyauInf > noyauSup) {
			throw new IntervalleMalDefiniException(
			"Ensemble des valeurs vraisemblables mal défini\n" +
			"Veuillez convenablement spécifier les bornes inférieure et supérieure");
		}
		// le noyau doit être inclus dans le support
		if (!( (supportInf <= noyauInf) && (noyauSup <= supportSup) )) {
			throw new NombreFlouMalDefiniException();
		}
				
		this.supportInf = supportInf;
		this.supportSup = supportSup;
		this.noyauInf = noyauInf;
		this.noyauSup = noyauSup;		
		this.incertitude = incertitude;
		this.desc = description;
		
	}

	/**
	 * Retourne une alpha-coupe
	 * @param alpha le niveau de l'alpha-coupe
	 * @return l'alpha-coupe (union d'intervalles)
	 */
	public Vector<Intervalle> getAlphaCoupe(Double alpha) {
		
		/*
		 * Élimination des niveaux de coupe incorrects
		 */
		if (alpha.compareTo(0d) < 0 || alpha.compareTo(1d) > 1) {
			return null;
		}
		
		Double valeur_noyau = 0d; // le niveau de possibilité des valeurs du noyau
		Intervalle coupeEnZero = null;
		try {
			switch (getIncertitude()) {
			
			// on sépare tous les cas pour plus de souplesse en cas de modification
			
			case SUR:
				valeur_noyau = 1d;
				break;
				
			case RAISONNABLE:
				valeur_noyau = 1d;
				if (alpha.compareTo(0.5d) <= 0) // alpha <= 0.5
					coupeEnZero = new Intervalle(0d, 0d);
				break;
			
			case INCERTAIN:
				valeur_noyau = 1d;
				//if (alpha <= 1d)
					coupeEnZero = new Intervalle(0d, 0d);
				break;
			
			case PLUTOT_INCERTAIN:
				valeur_noyau = 0.6d;
				//if (alpha <= 1d)
					coupeEnZero = new Intervalle(0d, 0d);
				break;
			
			case TRES_INCERTAIN:
				valeur_noyau = 0.2d;
				//if (alpha <= 1d)
					coupeEnZero = new Intervalle(0d, 0d);
				break;
	
			}
		} catch (IntervalleMalDefiniException e) {
			/* 
			 * ne devrait jamais survenir ici...
			 * on a défini les intervalles (proprement) dans le code
			 */
			//System.out.println(e.getMessage());
			e.printStackTrace();
		}

		// morceau du trapèze
		Intervalle coupeTrapeze = null;

		if (alpha.compareTo(valeur_noyau) <= 0) { // alpha <= valeur_noyau
			// les bornes de la coupe du trapèze
			Double borneInf, borneSup;

			/*
			 * NB : dans ce qui suit on n'a jamais valeur_noyau == 0
			 * et donc les pentes ne sont jamais nulles (pas de division par 0)
			 */
			if (getSupportInf().compareTo(getNoyauInf())==0) { 
				borneInf = getSupportInf();
			} else {		
				Double penteGauche = valeur_noyau / (getNoyauInf() - getSupportInf());
				Double decalageGauche = ( valeur_noyau * getSupportInf() ) / (getSupportInf() - getNoyauInf());
				borneInf = (alpha-decalageGauche)/penteGauche;
			}

			if (getNoyauSup().compareTo(getSupportSup())==0) {
				borneSup = getSupportSup();			
			} else {
				Double penteDroite= valeur_noyau / (getNoyauSup() - getSupportSup()); //négatif
				Double decalageDroite = ( valeur_noyau * getSupportSup() ) / (getSupportSup() - getNoyauSup());
				borneSup = (alpha-decalageDroite)/penteDroite;
			}

			try {
				coupeTrapeze = new Intervalle(borneInf, borneSup);
			} catch (IntervalleMalDefiniException e) {
				// même chose que ci-dessus
				//System.out.println(e.getMessage());
				e.printStackTrace();
			}
		}

		Vector<Intervalle> coupe = new Vector<Intervalle>();
		if (coupeEnZero != null) {
			coupe.add(coupeEnZero);
		}
		if (coupeTrapeze != null) {
			// la méthode "insererDansUnion" gère d'elle-même les problèmes de recouvrement
			coupeTrapeze.insererDansUnion(coupe);
		}
		
		Collections.sort(coupe);
		
		return coupe;
	}
	
	/*
	 * Getters
	 * Setters de l'incertitude et de la description
	 */
	
	public Double getSupportInf() {
		return supportInf;
	}

	public Double getSupportSup() {
		return supportSup;
	}

	public Double getNoyauInf() {
		return noyauInf;
	}

	public Double getNoyauSup() {
		return noyauSup;
	}

	public Incertitude getIncertitude() {
		return incertitude;
	}
	
	public void setIncertitude(Incertitude incertitude) {
		this.incertitude = incertitude;
	}

	public String getDesc() {
		return desc;
	}

	public void setDesc(String desc) {
		this.desc = desc;
	}
	
	public String toString() {
		String resultat = "";
		resultat += "\nNom : "+this.desc;
		resultat += "\nNoyauInf : "+this.noyauInf;
		resultat += "\nNoyauSup : "+this.noyauSup;
		resultat += "\nSupportInf : "+this.supportInf;
		resultat += "\nSupportSup : "+this.supportSup;
		resultat += "\nCertitude : "+this.incertitude;
		return resultat;
		
	}
	
}
