package estimateur.modele;

import java.util.Vector;



/**
 * Type intervalle
 * 
 *  Permet de représenter un intervalle réel et d'effectuer la somme avec une union
 *  d'autres intervalles
 */
public class Intervalle implements Comparable<Intervalle> {
	
	/*
	 * Les bornes de l'intervalle
	 */
	private Double inf, sup;
	
	/**
	 * Constructeur
	 * @param inf: borne inférieure de l'intervalle
	 * @param sup: borne supérieure de l'intervalle
	 */
	public Intervalle(Double inf, Double sup) throws IntervalleMalDefiniException {
		/*
		 * On vérifie la cohérence de la définition
		 * NB: on autorise un intervalla réduit à un seul élément
		 */
		if (inf > sup) {
			throw new IntervalleMalDefiniException("Bornes inf > borne sup");
		}
		
		this.inf = inf;
		this.sup = sup;
	}
	
	public Double getInf() {
		return inf;
	}

	public Double getSup() {
		return sup;
	}

	/**
	 * Détermine si l'intervalle a une intersection avec i
	 * @param i un intervalle
	 * @return vrai ssi l'intervalle et i ont une intersection non nulle
	 */
	public boolean intersecte(Intervalle i) {
		/*
		 * Il y a intersection si une, au moins, des bornes de l'intervalle est
		 * comprise entre celles de i	
		 */
		return ( getInf() >= i.getInf() && getInf() <= i.getSup() ) ||
			( getSup() >= i.getInf() && getSup() <= i.getSup() );
	}
	
	/**
	 * Insère "proprement" (en gérant les intersections) l'intervalle dans une union
	 * @param union l'union à laquelle ajouter l'intervalle
	 */
	public void insererDansUnion(Vector<Intervalle> union) {
		/*
		 * On va gérer les intervalles deux à deux : pour chacun des intervalles de
		 * l'union, on étudie les intersections
		 * s'il n'y en a pas, on passe au suivant, et s'il n'y a jamais intersection,
		 * 	l'intervalle est ajouté à l'union
		 * s'il y a intersection, on réalise l'union de l'intervalle avec celui qui
		 * 	l'intersecte, et on ajoute alors cet unique intervalle à l'union, récursivement
		 */
		int ind = 0;
		while (ind < union.size() && !intersecte(union.get(ind))) {ind++;}
		
		if (ind == union.size()) { // aucune intersection
			union.add(this);
		} else {
			// intersection avec union.get(ind)
			Intervalle i = union.get(ind);
			// on considère l'intervalle union de this et de union.get(ind)
			Intervalle fusion = null;
			try {
				fusion = new Intervalle( Math.min(getInf(), i.getInf()),
						Math.max(getSup(), i.getSup()));
			} catch (IntervalleMalDefiniException e) {
				e.printStackTrace();
			}
			// on retire i de l'union, car il est inclus dans fusion
			union.removeElement(i);
			// on ajoute récursivement fusion dans l'union
			/*
			 * NB : terminaison du fait qu'on a retiré un élément
			 */
			fusion.insererDansUnion(union);
		}
	}
	
	/**
	 * Réalise la somme de l'intervalle avec l'union d'intervalles donnée en paramètre
	 * @param union l'union d'intervalles avec laquelle on veut faire la somme
	 * @return l'union "propre" (recouvrements gérés) d'intervalles résultante
	 */
	public Vector<Intervalle> ajouter(Vector<Intervalle> union) {
		
		if (union==null)
			return null;
		
		Vector<Intervalle> resultat = new Vector<Intervalle>();
		
		for (Intervalle i : union) {
			Intervalle somme = null;
			try {
				// somme de l'intervalle avec i
				somme = new Intervalle(getInf()+i.getInf(), getSup()+i.getSup());
			} catch (IntervalleMalDefiniException e) {
				e.printStackTrace();
			}
			// insertion "propre" dans l'union résultat
			somme.insererDansUnion(resultat);
		}
		
		return resultat;
	}
	
	/**
	 * Savoir si un nombre appartient ou non à l'intervalle.
	 * @param d nombre à tester
	 * @return vrai si d appartient à l'intervalle
	 */
	
	public boolean contient(double d) {
		return ((d >= getInf()) && (d <= getSup()));
	}
	
	/** Seulement pour le graphique :
	 * Savoir quels sont les intervalles d'un ensemble qui 
	 * ont une intersection non-nulle avec cet intervalle. 
	 * @param vect intervalles qui sont soit entièrement inclus soit pas du tout
	 * @return les intervalles d'intersection non nulle avec l'intervalle de base
	 */
	public Vector<Intervalle> sousMembres(Vector<Intervalle> vect){
		Vector<Intervalle> resultat = new Vector<Intervalle>();
		for (Intervalle i : vect){
			if (contient(i.getInf())) resultat.add(i);
		}
		return resultat;
	}
	
	public String toString() {
		return "[ " + getInf() + " , " + getSup() + " ]";
	}

	public String pourLaPhrase() {
		return "from " + getInf().floatValue() + "€ to " + getSup().floatValue() + "€";
	}
	
	/**
	 * Comparaison entre intervalles, non pas au sens de l'inclusion ici, mais
	 * de la borne inférieure, en vue du tracé. Comme cette méthode sera appliquée
	 * à des unions "propres" (sans recouvrements), i et this ne sont jamais
	 * en inclusion.
	 * @param i un intervalle
	 */
	public int compareTo(Intervalle i) {
		return getInf().compareTo(i.getInf());
	}
	
}
