package estimateur.modele;

import java.util.Collections;
import java.util.HashMap;
import java.util.Vector;

/**
 * Représente la somme de sources floues de financement
 */
public class Somme {

	// on va sommer les alpha-coupes par pas de pas_alpha
	public static Double pas_alpha = 0.01d;
	
	// les alpha-coupes constituant la somme, par pas de pas_alpha
	private HashMap<Float, Vector<Intervalle>> alpha_coupes;
	
	/**
	 * Génère la somme de financements flous fournis
	 * @param financements les financements flous à sommer
	 */
	public Somme(Vector<Financement> financements) {
		
		alpha_coupes = new HashMap<Float, Vector<Intervalle>>();
		
		// on fait la somme des alpha coupes des financements par pas de pas_alpha
		Double alpha = 0d;
		while (alpha.compareTo(1d)<=0) { //alpha <= 1d
			// le vecteur sera modifié !
			Vector<Financement> copieDesFinancements = (Vector<Financement>)financements.clone();
			Vector<Intervalle> alpha_coupe = sommerAlphaCoupes(alpha, copieDesFinancements);
			Collections.sort(alpha_coupe);
			alpha_coupes.put(alpha.floatValue(), alpha_coupe);
			alpha += pas_alpha;
		}
		
	}
	
	/**
	 * Somme entre elles les alpha-coupes de même niveau alpha de plusieurs sources de financement
	 * @param alpha le niveau des alpha-coupes
	 * @param financements les financements dont on va prendre les alpha-coupes
	 * @return l'union d'intervalles somme des alpha-coupes
	 */
	public Vector<Intervalle> sommerAlphaCoupes(Double alpha, Vector<Financement> financements) {
		
		// si ensemble vide de financements...
		if (financements.size()==0)
			return null;
			
		// schéma récursif
		if (financements.size()==1) { // cas terminal : financement unique
			return financements.firstElement().getAlphaCoupe(alpha); // ensemble vide...
		} else { // cas général
			Vector<Intervalle> premieresAlphaCoupes = financements.firstElement().getAlphaCoupe(alpha);
			// somme récursive des autres financements
			financements.removeElementAt(0);
			Vector<Intervalle> sommeDuReste = sommerAlphaCoupes(alpha, financements);
			// ajout à la somme récursive des autres financements
			Vector<Intervalle> somme = new Vector<Intervalle>();
			for (Intervalle i : premieresAlphaCoupes) {
				fusionner(somme, i.ajouter(sommeDuReste));
			}
			return somme;
		}
				
	}
	
	// auxiliaire pour la méthode précédente
	/*
	 * Réalise l'union des deux unions passées en paramètre dans la première
	 */
	private void fusionner(Vector<Intervalle> union1, Vector<Intervalle> union2) {
		for (Intervalle i : union2) {
			i.insererDansUnion(union1);
		}
	}
	
	/**
	 * Exprime de manière textuelle le résultat de la somme des financements.
	 * @param html indique si on doit utiliser ou non <br /> au lieu de \n
	 * @return une chaîne décrivant la somme des financements.
	 */
	public String estimer(boolean html) throws ErreurEstimationException {
		String result = html ? "<html>" : "";
		String nl = html ? "<br/>" : "\n";
		
		/*
		 * on récupère les différents événements possibles
		 * et pour chacun on recherche le plateau le plus élevé (valeurs préférées)
		 */
		Vector<Intervalle> evenements = (Vector<Intervalle>)getAlphaCoupe(0d).clone(); // support
		// évévenements ordonnés selon le niveau alpha de leurs valeurs préférées, pour générer la phrase
		Vector<Evenement> evenementsOrdonnes = new Vector<Evenement>();
		// on map des d'unions : on peut avoir des coupes dinstinctes à un même niveau, pour un même événement
		HashMap<Intervalle, Vector<Intervalle>> preferees = new HashMap<Intervalle, Vector<Intervalle>>();
		for (Intervalle i : evenements) {
			preferees.put(i, new Vector<Intervalle>());
		}
		
		Double alpha = 1d; // recherche par le haut

		// vecteur donnant, pour chaque itération, les événements dont on a trouvé
		// les valeurs préférées, et qu'il faut donc supprimer
		Vector<Intervalle> evenementsSatisaits = new Vector<Intervalle>();

		// recherche des valeurs préférées pour chaque événement
		while (alpha.compareTo(0d)>0 && evenements.size()>0) { // alpha > 0 && ...
			
			evenementsSatisaits.clear();
			
			Vector <Intervalle> coupes = getAlphaCoupe(alpha);
			// on recherche les intersections avec le support, car si une alpha-coupe
			// d'un niveau plus grand qu'une autre l'intersecte, elle y est incluse
			// (il y a emboîtement vers le bas)
			if (coupes != null) {
				
				for (Intervalle coupe : coupes) {
					
					int ind = 0;
					while (ind < evenements.size() && !coupe.intersecte( evenements.get(ind) )) {
						//System.out.println("alpha:" + alpha  + " coupe: "+coupe+"  event: "+ evenements.get(ind));
						ind++;
					}
					
					if (ind < evenements.size()) { // si on a trouvé, normalement oui
						preferees.get( evenements.get(ind) ).add(coupe);
						evenementsSatisaits.add( evenements.get(ind) );
					}
					
				}
				
				// nettoyage et tri
				for (Intervalle i : evenementsSatisaits) {
					evenements.remove(i);
					evenementsOrdonnes.add( new Evenement(i, alpha) );
				}
				
			} // if (coupes != null)
			
			alpha -= pas_alpha;
		} // while (alpha.compareTo(0d)>0 && evenements.size()>0)
		
		if (evenements.size() > 0) {
			throw new ErreurEstimationException();
		}
		
		// génération de la chaîne
		result += "The calculation on the funding sources shows that:\n" +
			"(from the most possible to the least possible)"+nl;
		for (Evenement e : evenementsOrdonnes) {
			Double alpha_preferees = e.getAlpha_preferees();
			
			// texte modulé selon la possibilité
			if (alpha_preferees.compareTo(1d)==0) {
				result += "- it is absolutely possible to get ";
			} else if (alpha_preferees.compareTo(0.8d)>=0) { // et < 1
				result += "- it is quite possible to get ";
			} else if (alpha_preferees.compareTo(0.6d)>=0) { // et < 0.8
				result += "- it is uncertain to get ";
			} else if (alpha_preferees.compareTo(0.4d)>=0) { // et < 0.6
				result += "- it is unlikely to get ";
			} else if (alpha_preferees.compareTo(0.2d)>=0) { // et < 0.4
				result += "- it is very unlikely to get ";
			} else {
				result += "- it is almost impossible to get ";
			}
				
			result += decrirePreferees( preferees.get(e.getEvenement()) ) + 
				", within a range " +
				e.getEvenement().pourLaPhrase() + nl;
			
		}
		
		if (html) result += "</html>";
		
		return result;
	}
	
	// helper pour générer la phrase
	private String decrirePreferees(Vector<Intervalle> preferees) {
		String desc = "";
		
		for (int i=0; i<preferees.size()-1; i++) {
			desc += preferees.get(i).pourLaPhrase();
			desc += " ou ";
		}
		desc += preferees.lastElement().pourLaPhrase();
		
		return desc;
	}
	
	public Vector<Intervalle> getAlphaCoupe(Double alpha) {
		return alpha_coupes.get(alpha.floatValue());
	}
	
	public HashMap<Float, Vector<Intervalle>> getAlphaCoupes() {
		return this.alpha_coupes;
	}
	
}
